/*
 * $Id: mkinl.h,v 1.11 2003/11/17 21:50:55 spoel Exp $
 * 
 *                This source code is part of
 * 
 *                 G   R   O   M   A   C   S
 * 
 *          GROningen MAchine for Chemical Simulations
 * 
 *                        VERSION 3.2.0
 * Written by David van der Spoel, Erik Lindahl, Berk Hess, and others.
 * Copyright (c) 1991-2000, University of Groningen, The Netherlands.
 * Copyright (c) 2001-2004, The GROMACS development team,
 * check out http://www.gromacs.org for more information.

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * If you want to redistribute modifications, please consider that
 * scientific software is very special. Version control is crucial -
 * bugs must be traceable. We will be happy to consider code for
 * inclusion in the official distribution, but derived work must not
 * be called official GROMACS. Details are found in the README & COPYING
 * files - if they are missing, get the official version at www.gromacs.org.
 * 
 * To help us fund GROMACS development, we humbly ask that you cite
 * the papers on the package - you can find them in the top README file.
 * 
 * For more info, check our website at http://www.gromacs.org
 * 
 * And Hey:
 * GROningen Mixture of Alchemy and Childrens' Stories
 */

#ifndef _mkinl_h
#define _mkinl_h

/* This file is NOT threadsafe, but it is only used to create
 * the innerloops during the build process, so it will never be
 * executed by multiple threads.
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <types/simple.h>
#include <metacode.h>

#define STRINGSIZE 1024

#define WATERATOMS 3

/* Most options are controlled by these defines.
 * There are two major reasons for this:
 * First, it makes the code somewhat less cluttered.
 * Second, it makes it easier to introduce new interactions and/or
 * change the code without changing all if statements - just set
 * the options in mkinl.c and most things should follow automatically.
 */

#define DO_FORCE       (loop.do_force==TRUE)
#define DO_RF          (loop.coul==COUL_RF)
#define DO_COULTAB     (loop.coul==COUL_TAB)
#define DO_VDWTAB       ((loop.vdw==VDW_TAB) || (loop.vdw==VDW_BHAMTAB))
#define DO_TAB         (DO_COULTAB || DO_VDWTAB)
#define DO_BHAM        ((loop.vdw==VDW_BHAM) || (loop.vdw==VDW_BHAMTAB))
#define DO_PREFETCH    ((loop.sol==SOL_NO && (opt.prefetch & TYPE_NORMAL)) || \
                        (loop.sol==SOL_MNO && (opt.prefetch & TYPE_SOLVENT)) || \
                        (loop.sol==SOL_WATER && (opt.prefetch & TYPE_WATER)) || \
                        (loop.sol==SOL_WATERWATER && (opt.prefetch & TYPE_WATERWATER)))
  
#define DO_SOL         (loop.sol==SOL_MNO) /* non-water solvent optimzations */
#define DO_WATER       (loop.sol==SOL_WATER || loop.sol==SOL_WATERWATER)
#define DO_SOFTCORE    (loop.free==FREE_SOFTCORE)

#define DO_VECTORIZE    (loop.vectorize_invsqrt || loop.vectorize_recip)

/* For pure LJ-only loops we can save some cycles by just calculating the
 * reciprocal. In all other cases we need to do the invsqrt
 */
#define DO_INLINE_INVSQRT   (loop.invsqrt && !loop.vectorize_invsqrt && opt.inline_invsqrt)
#define OVERWRITE_RSQ  (!arch.vectorcpu && !loop.vdw_needs_r && !loop.vdw_needs_rsq && \
                        !loop.coul_needs_r && !loop.coul_needs_rsq)
			/* Can we overwrite the vectorization array
			 * holding rsq with the resulting rinv or rinvsq,
			 * or do we need to keep rsq a while?
			 */
#define N_VDWPARAM       (DO_BHAM ? 3 : 2)
			
/* Enumerations for different types of innerloops. These might be similar,
 * but not identical, to some of the definitions in include/types/enums.h.
 * Here we also need to take into account whether we use tables or not,
 * and whether a certain innerloop should calculate free energy or not.
 */


/* Electrostatics */
typedef enum {
  COUL_NO,
  COUL_NORMAL,          /* ordinary coulomb, constant epsilon */
  COUL_RF,              /* reaction field */
  COUL_TAB,             /* tabulated interactions */
  COUL_NR
} coul_t;

/* Alternatives for nonbonded interactions */
typedef enum {
  VDW_NO,
  VDW_LJ,              /* Lennard-Jones 6-12 interactions */
  VDW_BHAM,            /* Buckingham */
  VDW_TAB,             /* tabulated interactions */
  VDW_BHAMTAB,         /* Exponential tab, for buckingham interactions */
  VDW_NR
} vdw_t;

typedef enum {
  SOL_NO,             /* no water or solvent optimization */
  SOL_MNO,            /* interactions with solvent consisting of N
                       * lj+coul atoms and M atoms with coul only */ 
  SOL_WATER,          /* interactions with a 3-atom water molecule */
  SOL_WATERWATER,     /* interactions between two water molecules */
  SOL_NR
} sol_t;

typedef enum {
  FREE_NO,
  FREE_LAMBDA,      /* A separate free energy innerloop (lambda)      */
  FREE_SOFTCORE,    /* is called for alpha=0 or lambda=0 or lambda=1, */ 
  FREE_NR           /* to avoid if statements or extra table lookups. */
} free_t;


typedef enum {
  THREAD_NO,
  THREAD_MUTEX,
  THREAD_STRIDE
} thread_t;


/* This is essentially a boolean telling for which loop types
 * some features are enabled.
 * Bit 0 concerns normal loops,
 * bit 1 solvent loops and
 * bit 2 water loops and
 * bit 3 water-water loops.
 */
typedef int looptype_t;

#define TYPE_NORMAL     1
#define TYPE_SOLVENT    2
#define TYPE_WATER      4
#define TYPE_WATERWATER 8


/* Global structure determining architectural options */
typedef struct {
  bool     gmx_invsqrt;    /* Use gmx software invsqrt? */
  thread_t threads;
  bool     simplewater;
  bool     vectorcpu;          
  bool     cray_pragma;
} arch_t;


/* Global structure determining optimization options */
typedef struct {
  bool         inline_invsqrt;            /* Inline gmx software invsqrt */
  looptype_t   prefetch;
  bool         decrease_lookup_latency;   /* try to hide invsqrt lookup penalty */
  bool         delay_invsqrt;
  looptype_t   vectorize_invsqrt;         /* vectorize distance calculation when possible */
  bool         vectorize_recip;           /* vector recip is not meaningful on water loops */
} opt_t;


/* Global structure determining options for the individual
 * loop currently being written
 */ 

typedef struct {
    coul_t   coul;         /* What kind of loop is this? */
    vdw_t    vdw;
    sol_t    sol;
    free_t   free;
    bool     do_force;           /* dont calc force in MC loop versions */
    bool     coul_needs_rinv;    /* Which power of r are needed for the */
    bool     coul_needs_rinvsq;  /* coulombic interactions?             */
    bool     coul_needs_rsq;
    bool     coul_needs_r;
    bool     vdw_needs_rinv;    /* And for the nonbonded? */
    bool     vdw_needs_rinvsq;
    bool     vdw_needs_rsq;
    bool     vdw_needs_r;
    bool     invsqrt;           /* tells what we are calculating */
    bool     recip;
    bool     vectorize_invsqrt; 
    bool     vectorize_recip;
    int      ni;
    int      nj;
} loop_t;


extern arch_t arch;
extern opt_t  opt;
extern loop_t loop;
extern bool writeback;

extern char loopname[STRINGSIZE];

extern int table_element_size;

extern char forcebuf[255];
extern char tabforcebuf[255];

/* Exported routines from each file */
 
/* mkinl_declarations.c */
void file_header(void);
/* Adds include files for c, and calls fortran functions to be written
 * at the top of the source file
 */
void func_header(char *appendname);
/* Writes the header with some info and assigns the function name
 * to each loop
 */
void func_args(void);
/* Declares function arguments to the variable buffer */
void func_localvars(void);
/* Declares function variables to the buffer */
void func_init_vars(void);
/* Initiates some local data before the outermost loop */


/* Outer loop routines found in mkinl_outerloop.c
 */
void outer_loop(void);
/* Called to construct the loop over i particles, and in some
 * cases also loops over e.g. threads.
 */


/* Innerloop routines found in mkinl_innerloop.c
 */
void inner_loop(bool calcdist, bool calcforce);
/* Constructs the loop over j particles in a neighbour list */

/* Invsqrt calculation, mkinl_invsqrt.c
 */
int calc_invsqrt(void);
void invsqrt_vars(void);
void fortran_invsqrt(void);
int calc_recip(void);
void fortran_vecinvsqrt(void);
void fortran_vecsqrt(void);


/* The interactions, mkinl_interactions.c
 */

void start_stripmine_loop(void);
void close_stripmine_loop(void);

void call_vectorized_routines(void);

int calc_interactions(void);

int calc_dist(void);
int calc_rinv_and_rinvsq(void);

int calc_rsquare(char *atom1, char *atom2);

void prefetch_forces(void);

void unpack_inner_data(bool calcdist, bool calcforce);

void fetch_coord(void);

int update_inner_forces(int i,int j);

#endif
